/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.codehaus.activemq.service.impl;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.activemq.broker.Broker;
import org.codehaus.activemq.message.ActiveMQXid;
import org.codehaus.activemq.service.Transaction;
import org.codehaus.activemq.store.PreparedTransactionStore;
import org.codehaus.activemq.util.SerializationHelper;

import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.Map;

/**
 * @version $Revision: 1.3 $
 */
public class XATransactionCommand extends AbstractTransaction {
    private static final Log log = LogFactory.getLog(TransactionManagerImpl.class);

    private ActiveMQXid xid;
    private transient Map xaTxs;
    private transient PreparedTransactionStore preparedTransactions;

    /**
     * This constructor is only used when deserializing from disk
     */
    public XATransactionCommand() {
        super(null);
    }

    public XATransactionCommand(Broker broker, ActiveMQXid xid, Map xaTxs, PreparedTransactionStore preparedTransactions) {
        super(broker);
        this.xid = xid;
        this.xaTxs = xaTxs;
        this.preparedTransactions = preparedTransactions;
    }


    /**
     * Called after the transaction command has been recovered from disk
     *
     * @param xaTxs
     * @param preparedTransactions
     */
    public void initialise(Map xaTxs, PreparedTransactionStore preparedTransactions) {
        this.xaTxs = xaTxs;
        this.preparedTransactions = preparedTransactions;
    }

    public void commit(boolean onePhase) throws XAException {
        switch (getState()) {
            case START_STATE:
                // 1 phase commit, no work done.
                if (!onePhase) {
                    XAException xae = new XAException("Cannot do 2 phase commit if the transaction has not been prepared.");
                    xae.errorCode = XAException.XAER_PROTO;
                    throw xae;
                }
                setState(AbstractTransaction.FINISHED_STATE);
                xaTxs.remove(xid);
                break;
            case IN_USE_STATE:
                if (!onePhase) {
                    XAException xae = new XAException("Cannot do 2 phase commit if the transaction has not been prepared.");
                    xae.errorCode = XAException.XAER_PROTO;
                    throw xae;
                }
                // 1 phase commit, work done.
                try {
                    prePrepare();
                }
                catch (XAException e) {
                    throw e;
                }
                catch (Throwable e) {
                    log.warn("COMMIT FAILED: ", e);
                    rollback();

                    XAException xae = new XAException("ONE PHASE COMMIT FAILED: Transaction rolled back.");
                    xae.errorCode = XAException.XA_RBOTHER;
                    xae.initCause(e);
                    throw xae;
                }

                setState(AbstractTransaction.FINISHED_STATE);
                xaTxs.remove(xid);

                try {
                    postCommit();
                }
                catch (Throwable e) {
                    // I guess this could happen.  Post commit task failed
                    // to execute properly.
                    log.warn("POST COMMIT FAILED: ", e);
                    XAException xae = new XAException("POST COMMIT FAILED");
                    xae.errorCode = XAException.XAER_RMERR;
                    xae.initCause(e);
                    throw xae;
                }
                break;

            case PREPARED_STATE:
                // 2 phase commit, work done.
                // We would record commit here.
                setState(AbstractTransaction.FINISHED_STATE);
                xaTxs.remove(xid);

                try {
                    postCommit();
                }
                catch (Throwable e) {
                    // I guess this could happen.  Post commit task failed
                    // to execute properly.
                    log.warn("POST COMMIT FAILED: ", e);
                    XAException xae = new XAException("POST COMMIT FAILED");
                    xae.errorCode = XAException.XAER_RMERR;
                    xae.initCause(e);
                    throw xae;
                }
                break;
            default:
                XAException xae = new XAException("Cannot call commit now.");
                xae.errorCode = XAException.XAER_PROTO;
                throw xae;
        }
    }

    public void rollback() throws XAException {

        switch (getState()) {
            case START_STATE:
                // 1 phase rollback no work done.
                xaTxs.remove(xid);
                setState(AbstractTransaction.FINISHED_STATE);

                break;
            case IN_USE_STATE:
                // 1 phase rollback work done.
                xaTxs.remove(xid);
                setState(AbstractTransaction.FINISHED_STATE);

                try {
                    postRollback();
                }
                catch (Throwable e) {
                    // I guess this could happen.  Post commit task failed
                    // to execute properly.
                    log.warn("POST ROLLBACK FAILED: ", e);
                    XAException xae = new XAException("POST ROLLBACK FAILED");
                    xae.errorCode = XAException.XAER_RMERR;
                    xae.initCause(e);
                    throw xae;
                }

                break;
            case PREPARED_STATE:
                // 2 phase rollback work done.
                preparedTransactions.remove(xid);
                xaTxs.remove(xid);
                setState(AbstractTransaction.FINISHED_STATE);

                try {
                    postRollback();
                }
                catch (Throwable e) {
                    // I guess this could happen.  Post commit task failed
                    // to execute properly.
                    log.warn("POST ROLLBACK FAILED: ", e);
                    XAException xae = new XAException("POST ROLLBACK FAILED");
                    xae.errorCode = XAException.XAER_RMERR;
                    xae.initCause(e);
                    throw xae;
                }
                break;
        }

    }

    public int prepare() throws XAException {
        switch (getState()) {
            case START_STATE:
                // No work done.. no commit/rollback needed.
                xaTxs.remove(xid);
                setState(AbstractTransaction.FINISHED_STATE);
                return XAResource.XA_RDONLY;

            case IN_USE_STATE:
                try {
                    prePrepare();
                }
                catch (XAException e) {
                    throw e;
                }
                catch (Throwable e) {
                    log.warn("PREPARE FAILED: ", e);
                    rollback();

                    // Let them know we rolled back.
                    XAException xae = new XAException("PREPARE FAILED: Transaction rollback.");
                    xae.errorCode = XAException.XA_RBOTHER;
                    xae.initCause(e);
                    throw xae;
                }

                // We would record prepare here.
                setState(AbstractTransaction.PREPARED_STATE);
                preparedTransactions.put(xid, this);

                return XAResource.XA_OK;
            default :
                XAException xae = new XAException("Cannot call prepare now.");
                xae.errorCode = XAException.XAER_PROTO;
                throw xae;
        }
    }

    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal(in);
        xid = ActiveMQXid.read(in);
    }

    public void writeExternal(ObjectOutput out) throws IOException {
        super.writeExternal(out);
        xid.writeExternal(out);
    }

    public static Transaction fromBytes(byte[] data) throws IOException, ClassNotFoundException {
        return (Transaction) SerializationHelper.deserialize(data);
    }

    public byte[] toBytes() throws IOException {
        return SerializationHelper.serialize(this);
    }
}